@@ -28,14 +28,13 @@
 
 #include "md5.h"
 #include "support.h"
+#include "bigcrypt.h"
 #ifdef WITH_SELINUX
 #include <selinux/selinux.h>
 #define SELINUX_ENABLED is_selinux_enabled()>0
 #else
 #define SELINUX_ENABLED 0
 #endif
-extern char *crypt(const char *key, const char *salt);
-extern char *bigcrypt(const char *key, const char *salt);
 
 /* this is a front-end for module-application conversations */
 
@@ -527,12 +526,12 @@ static int _unix_run_helper_binary(pam_handle_t *pamh, const char *passwd,
 	}
 
 	/* exec binary helper */
-	args[0] = x_strdup(CHKPWD_HELPER);
+	args[0] = strdup(CHKPWD_HELPER);
 	args[1] = x_strdup(user);
 	if (off(UNIX__NONULL, ctrl)) {	/* this means we've succeeded */
-	  args[2]=x_strdup("nullok");
+	  args[2]=strdup("nullok");
 	} else {
-	  args[2]=x_strdup("nonull");
+	  args[2]=strdup("nonull");
 	}
 
 	execve(CHKPWD_HELPER, args, envp);
@@ -566,7 +565,7 @@ static int _unix_run_helper_binary(pam_handle_t *pamh, const char *passwd,
 	retval = PAM_AUTH_ERR;
     }
 
-    if (sighandler != NULL) {
+    if (sighandler != SIG_ERR) {
         (void) signal(SIGCHLD, sighandler);   /* restore old signal handler */
     }
 
@@ -690,15 +689,23 @@ int _unix_verify_password(pam_handle_t * pamh, const char *name
 		    D(("user has empty password - access denied"));
 		    retval = PAM_AUTH_ERR;
 		}
-	    } else if (!p || (*salt == '*') || (salt_len < 13)) {
+	    } else if (!p || (*salt == '*')) {
 		retval = PAM_AUTH_ERR;
 	    } else {
 		if (!strncmp(salt, "$1$", 3)) {
+		    salt_len = 0;
 		    pp = Goodcrypt_md5(p, salt);
 		    if (strcmp(pp, salt) != 0) {
 			_pam_delete(pp);
 			pp = Brokencrypt_md5(p, salt);
 		    }
+		} else if (*salt == '$') {
+                    /*
+		     * Ok, we don't know the crypt algorithm, but maybe
+		     * libcrypt nows about it? We should try it.
+		     */
+		    salt_len = 0;
+		    pp = x_strdup (crypt(p, salt));
 		} else {
 		    pp = bigcrypt(p, salt);
 		}
@@ -716,7 +723,8 @@ int _unix_verify_password(pam_handle_t * pamh, const char *name
 		 * stored string with the subset of bigcrypt's result.
 		 * Bug 521314: The strncmp comparison is for legacy support.
 		 */
-		if (strncmp(pp, salt, salt_len) == 0) {
+		if ((!salt_len && strcmp(pp, salt) == 0) || 
+		    (salt_len && strncmp(pp, salt, salt_len) == 0)) {
 		    retval = PAM_SUCCESS;
 		} else {
 		    retval = PAM_AUTH_ERR;
